<?php

namespace App\Http\Controllers\Api\V1;

use App\Http\Controllers\Controller;
use App\Http\Resources\RecipeResource;
use App\Http\Traits\HasStoreFile;
use App\Models\Recipe;
use App\Models\RecipeDetail;
use Illuminate\Http\Request;

class RecipeController extends Controller
{
    use HasStoreFile;

    // path of the folder where files related to the controller will be stored
    protected string $folderPath = "recipes";

    // name of the field for file upload in the model
    protected string $fileField = "image";

    public function index(Request $request) {
        $perPage = $request->perPage ?? 10;
        $recipes = Recipe::with(["franchise", "recipeDetails.item"])
            ->filter()->latest()->paginate($perPage);

        return RecipeResource::collection($recipes);
    }

    public function store(Request $request) {
        $data = $request->validate([
            'franchise_id' => ['required', "string"],
            'name' => ['required', "string"],
            'price' => ['required'],
            'gst' => ['sometimes'],
            'image' => ['sometimes', "file", "mimes:jpeg,jpg,png,gif", "max:2048"],
            "recipe_details" => ["sometimes", "array"]
        ]);

        $recipe = new Recipe($data);
        if ($request->hasFile($this->fileField)) {
            $recipe[$this->fileField] = $this->storeFile($request, $this->fileField, $this->folderPath);
        }
        $recipe->save();

        if ($request->has("recipe_details")) {
            foreach ($request->recipe_details as $recipe_detail) {
                $recipeDetail = new RecipeDetail($recipe_detail);
                $recipeDetail->recipe_id = $recipe->id;
                $recipeDetail->save();
            }
        }

        return new RecipeResource($recipe);
    }

    public function show(Recipe $recipe) {
        return new RecipeResource($recipe->loadMissing(["franchise", "recipeDetails.item"]));
    }

    public function update(Request $request, Recipe $recipe) {
        $data = $request->validate([
            'franchise_id' => ['sometimes', "string"],
            'name' => ['sometimes', "string"],
            'price' => ['sometimes'],
            'gst' => ['sometimes'],
            'image' => ['sometimes', "file", "mimes:jpeg,jpg,png,gif", "max:2048"],
            "recipe_details" => ["sometimes", "array"],
            "delete_recipe_detail_ids" => ["sometimes", "array"]
        ]);

        $recipe->fill($data);
        if ($request->hasFile($this->fileField)) {
            $this->deleteFile($this->folderPath, $recipe[$this->fileField]);
            $recipe[$this->fileField] = $this->storeFile($request, $this->fileField, $this->folderPath);
        }
        $recipe->save();

        if ($request->has("recipe_details")) {
            foreach ($request->recipe_details as $recipe_detail) {
                unset($recipe_detail->item);
                if (!isset($recipe_detail->id)) {
                    $recipeDetail = new RecipeDetail($recipe_detail);
                    $recipeDetail->recipe_id = $recipe->id;
                    $recipeDetail->save();
                }
                RecipeDetail::findOrFail($recipe_detail->id)->update($recipe_detail);
            }
        }

        if ($request->has("delete_recipe_detail_ids")) {
            RecipeDetail::whereIn("id", $request->delete_recipe_detail_ids)->delete();
        }

        return new RecipeResource($recipe);
    }

    public function destroy(Recipe $recipe) {
        RecipeDetail::where("recipe_id", $recipe->id)->delete();
        $recipe->delete();

        return response()->json(["message" => "Recipe deleted successfully!"]);
    }
}
